/**
* The MIT License (MIT)
* 
* Copyright (c) 2013 Retroswitch LLC
* 
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
* 
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
* 
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
* THE SOFTWARE.
*
*/

//
// Includes

#include "PaintClient.h"
#include "ConnectDialog.h"
#include "Canvas.h"
#include "ProfileManager.h"
#include "SettingsManager.h"
#include <QWidgetAction>
#include <QMessageBox>
#include <QFileDialog>

//
// PaintClient Methods

/******************************************************************************/
/* Constructor                                                                */
/******************************************************************************/

PaintClient::PaintClient(QWidget *parent, Qt::WFlags flags) :
    QMainWindow( parent, flags ), m_tcpSocket( 0 )
{
    m_ui.setupUi( this );
    statusBar()->addPermanentWidget( m_statusLabel = new QLabel );

    m_canvas = new Canvas;
    setCentralWidget( m_canvas );

    connect( m_ui.actionConnect, SIGNAL( triggered() ), this, SLOT( OnActionConnect() ) );
    connect( m_ui.actionDisconnect, SIGNAL( triggered() ), this, SLOT( OnActionDisconnect() ) );
    connect( m_ui.actionQuit, SIGNAL( triggered() ), this, SLOT( close() ) );
    connect( m_ui.actionClear, SIGNAL( triggered() ), this, SLOT( OnActionClear() ) );
    connect( m_ui.actionLoadImage, SIGNAL( triggered() ), this, SLOT( OnActionLoadImage() ) );
    connect( m_ui.menuBorderColor, SIGNAL( triggered( QAction* ) ), this, SLOT( OnActionBorderColor( QAction* ) ) );
    connect( m_ui.menuBackgroundColor, SIGNAL( triggered( QAction* ) ), this, SLOT( OnActionBackgroundColor( QAction* ) ) );
    connect( m_ui.menuDrawingColor, SIGNAL( triggered( QAction* ) ), this, SLOT( OnActionPenColor( QAction* ) ) );
    connect( m_canvas, SIGNAL( CellChanged( uint8, uint8, uint8 ) ), this, SLOT( OnCellChanged( uint8, uint8, uint8 ) ) );

    SetState( cStateDisconnected );
}

/******************************************************************************/
/* Destructor                                                                 */
/******************************************************************************/

PaintClient::~PaintClient()
{
}

/******************************************************************************/
/* OnActionConnect                                                            */
/******************************************************************************/

void PaintClient::OnActionConnect()
{
    ConnectDialog dlg( this, 0 );

    if( dlg.exec() == QDialog::Accepted )
    {
        ConnectToServer( SettingsManager::GetInstance()->GetPreferredHost(), SettingsManager::GetInstance()->GetPreferredPort() );
    }
}

/******************************************************************************/
/* OnActionDisconnect                                                         */
/******************************************************************************/

void PaintClient::OnActionDisconnect()
{
    m_tcpSocket->disconnectFromHost();
}

/******************************************************************************/
/* OnActionClear                                                              */
/******************************************************************************/

void PaintClient::OnActionClear()
{
    m_canvas->Clear();
    SendCommandMessage( cCommandIdClear );
}

/******************************************************************************/
/* OnActionLoadImage                                                          */
/******************************************************************************/

void PaintClient::OnActionLoadImage()
{
    QString filename = QFileDialog::getOpenFileName( this, "Choose Image", "", "Image Files (*.png *.jpg *.gif *.bmp)" );

    if( filename.length() )
    {
        QImage image( filename );

        if( !image.width() || !image.height() )
        {
            QMessageBox::warning( this, "Error", "Unable to open image.", QMessageBox::Ok );
            return;
        }
        
        QVector<QRgb> color_table;
        const Profile* profile = m_canvas->GetProfile();

        for( int i=0; i<profile->m_colors.count(); i++ )
        {
            color_table.push_back( profile->m_colors.at( i ).rgba() );
        }

        QImage scaled = image.scaled( QSize( profile->m_columns, profile->m_rows ) );
        QImage converted = scaled.convertToFormat( QImage::Format_Indexed8, color_table );

        for( int y=0; y<profile->m_rows; y++ )
        {
            const uint8* bits = converted.constScanLine( y );

            for( int x=0; x<profile->m_columns; x++ )
            {
                m_canvas->Plot( uint8( x ), uint8( y ), bits[x] );
            }

            SendScanlineMessage( uint8( y ), uint8( profile->m_columns ), bits );
        }
    }
}

/******************************************************************************/
/* OnActionBorderColor                                                        */
/******************************************************************************/

void PaintClient::OnActionBorderColor( QAction* action )
{
    m_canvas->SetBorderColor( uint8( action->data().toInt() ) );
    SendCommandMessage( cCommandIdSetColors, m_canvas->GetBorderColor(), m_canvas->GetBackgroundColor() );
}

/******************************************************************************/
/* OnActionBackgroundColor                                                    */
/******************************************************************************/

void PaintClient::OnActionBackgroundColor( QAction* action )
{
    m_canvas->SetBackgroundColor( uint8( action->data().toInt() ) );
    SendCommandMessage( cCommandIdSetColors, m_canvas->GetBorderColor(), m_canvas->GetBackgroundColor() );
}

/******************************************************************************/
/* OnActionPenColor                                                           */
/******************************************************************************/

void PaintClient::OnActionPenColor( QAction* action )
{
    m_canvas->SetDrawingColor( uint8( action->data().toInt() ) );
}

/******************************************************************************/
/* OnSocketConnected                                                          */
/******************************************************************************/

void PaintClient::OnSocketConnected()
{
    // The Flyer has accepted the network connection, so we send a "connect"
    // message packet with our protocol version. If it matches the protocol the
    // paint server was built with, we'll receive a full update in response
    // containing the entire canvas contents and current screen colors.
    // Otherwise, we'll be disconnected :)

    MessageHeader header;
    ConnectPayload payload;

    header.m_id = cMessageIdConnect;
    header.m_pad = 0;
    header.m_payloadLength = sizeof( ConnectPayload );

    payload.m_protocol = MESSAGE_PROTOCOL;

    m_tcpSocket->write( (const char*)&header, sizeof( MessageHeader ) );
    m_tcpSocket->write( (const char*)&payload, sizeof( ConnectPayload ) );

    SetState( cStateWaitingForUpdate );
}

/******************************************************************************/
/* OnSocketDisconnected                                                       */
/******************************************************************************/

void PaintClient::OnSocketDisconnected()
{
    SetState( cStateDisconnected );
    m_canvas->SetProfile( 0 );
    m_tcpSocket->deleteLater();
}

/******************************************************************************/
/* OnSocketReadReady                                                          */
/******************************************************************************/

void PaintClient::OnSocketReadReady()
{
    // Buffer the received data

    m_message.append( m_tcpSocket->readAll() );

    // Process any complete messages in our data buffer, pulling them off the
    // front of the buffer as we go.

    while( m_message.size() >= 4 )
    {
        MessageHeader* header = (MessageHeader*)m_message.data();
        int total_first_message_size = sizeof( MessageHeader ) + header->m_payloadLength;

        if( m_message.size() < total_first_message_size )
        {
            break;
        }

        switch( header->m_id )
        {
            case cMessageIdUpdate:
                HandleUpdateMessage( header, (const UpdatePayload*)(m_message.data() + sizeof( MessageHeader )) );
                break;
        }
    
        m_message.remove( 0, total_first_message_size );
    }
}

/******************************************************************************/
/* OnCellChanged                                                              */
/******************************************************************************/

void PaintClient::OnCellChanged( uint8 x, uint8 y, uint8 color )
{
    SendCommandMessage( cCommandIdPlot, uint8( x ), uint8( y ), color );
}

/******************************************************************************/
/* SetState                                                                   */
/******************************************************************************/

void PaintClient::SetState( State state )
{
    switch( state )
    {
        case cStateDisconnected:
            m_ui.menuBar->setEnabled( true );
            m_ui.actionConnect->setEnabled( true );
            m_ui.actionDisconnect->setEnabled( false );
            m_ui.actionClear->setEnabled( false );
            m_ui.actionLoadImage->setEnabled( false );
            m_ui.menuBackgroundColor->setEnabled( false );
            m_ui.menuBorderColor->setEnabled( false );
            m_ui.menuDrawingColor->setEnabled( false );
            m_statusLabel->setText( "Disconnected" );
            break;

        case cStateConnecting:
            m_ui.menuBar->setEnabled( false );
            m_statusLabel->setText( "Connecting..." );
            break;

        case cStateWaitingForUpdate:
            m_ui.menuBar->setEnabled( false );
            m_statusLabel->setText( "Receiving update..." );
            break;

        case cStateConnected:
            m_ui.menuBar->setEnabled( true );
            m_ui.actionConnect->setEnabled( false );
            m_ui.actionDisconnect->setEnabled( true );
            m_ui.actionClear->setEnabled( true );
            m_ui.actionLoadImage->setEnabled( true );
            m_statusLabel->setText( "Connected" );
            break;

        default:
            break;
    }
    
    repaint();
}

/******************************************************************************/
/* ConnectToServer                                                            */
/******************************************************************************/

void PaintClient::ConnectToServer( const QString& ipAddress, uint16 port )
{
    m_tcpSocket = new QTcpSocket( this );

    connect( m_tcpSocket, SIGNAL( connected() ), this, SLOT( OnSocketConnected() ) );
    connect( m_tcpSocket, SIGNAL( disconnected() ), this, SLOT( OnSocketDisconnected() ) );
    connect( m_tcpSocket,  SIGNAL( readyRead() ), this, SLOT( OnSocketReadReady() ) );

    SetState( cStateConnecting );
    m_tcpSocket->connectToHost( QHostAddress( ipAddress ), port );

    if( m_tcpSocket->waitForConnected( 5000 ) == false )
    {
        QMessageBox::information( this, tr( "Unable to connect to %1:%2." ).arg( ipAddress ).arg( port ), m_tcpSocket->errorString(), QMessageBox::Ok );
        m_tcpSocket->deleteLater();
        SetState( cStateDisconnected );
    }
}

/******************************************************************************/
/* HandleUpdateMessage                                                        */
/******************************************************************************/

void PaintClient::HandleUpdateMessage( const MessageHeader* header, const UpdatePayload* updatePayload )
{
    (void)header;

    const Profile* profile = ProfileManager::GetInstance()->GetProfile( (ProfileManager::Target)updatePayload->m_target );

    if( !profile )
    {
        return;
    }

    // Configure and update canvas with the received data
            
    m_canvas->SetProfile( profile, ((uint8*)updatePayload) + sizeof( UpdatePayload ) );
    m_canvas->SetBorderColor( updatePayload->m_borderColor );
    m_canvas->SetBackgroundColor( updatePayload->m_backgroundColor );
    m_canvas->SetDrawingColor( 1 );

    // Configure menu options based on the current profile

    m_ui.menuBorderColor->clear();
    m_ui.menuBackgroundColor->clear();
    m_ui.menuDrawingColor->clear();

    if( profile->m_supportsScreenColor )
    {
        GenerateColorMenu( m_ui.menuBorderColor, profile->m_colors, m_canvas->GetBorderColor() );
        GenerateColorMenu( m_ui.menuBackgroundColor, profile->m_colors, m_canvas->GetBackgroundColor() );

        m_ui.menuBackgroundColor->setEnabled( true );
        m_ui.menuBorderColor->setEnabled( true );
    }

    GenerateColorMenu( m_ui.menuDrawingColor, profile->m_colors, m_canvas->GetDrawingColor() );
    m_ui.menuDrawingColor->setEnabled( true );

    // We're now ready to interact

    SetState( cStateConnected );
}

/******************************************************************************/
/* SendCommandMessage                                                         */
/******************************************************************************/

void PaintClient::SendCommandMessage( uint8 commandId, uint8 arg0, uint8 arg1, uint8 arg2 )
{
    MessageHeader header;
    CommandPayload payload;

    header.m_id = cMessageIdCommand;
    header.m_pad = 0;
    header.m_payloadLength = sizeof( CommandPayload );

    payload.m_id = commandId;
    payload.m_arg[0] = arg0;
    payload.m_arg[1] = arg1;
    payload.m_arg[2] = arg2;

    m_tcpSocket->write( (const char*)&header, sizeof( MessageHeader ) );
    m_tcpSocket->write( (const char*)&payload, header.m_payloadLength );
}

/******************************************************************************/
/* SendScanlineMessage                                                        */
/******************************************************************************/

void PaintClient::SendScanlineMessage( uint8 row, uint8 length, const uint8* data )
{
    MessageHeader header;
    ScanlinePayload payload;

    header.m_id = cMessageIdScanline;
    header.m_pad = 0;
    header.m_payloadLength = sizeof( ScanlinePayload ) + length;

    payload.m_row = row;
    payload.m_length = length;

    m_tcpSocket->write( (const char*)&header, sizeof( MessageHeader ) );
    m_tcpSocket->write( (const char*)&payload, sizeof( ScanlinePayload ) );
    m_tcpSocket->write( (const char*)data, length );
}

/******************************************************************************/
/* GenerateColorMenu                                                          */
/******************************************************************************/

QActionGroup* PaintClient::GenerateColorMenu( QMenu* menu, const QList<QColor>& colors, int initialSelection )
{
    QActionGroup* group = new QActionGroup( menu );
    group->setExclusive( true );

    for( int i=0; i<colors.count(); i++ )
    {
        QAction* action = new QAction( menu );
        action->setText( colors.at( i ).name() );
        QPixmap pixmap( 10, 10 ); pixmap.fill( QColor( colors.at( i ) ) );
        action->setIcon( QIcon( pixmap ) );
        action->setCheckable( true );
        action->setData( i );
        action->setIconVisibleInMenu( true );
        group->addAction( action );

        if( i == initialSelection )
        {
            action->setChecked( true );
        }
    }

    menu->addActions( group->actions() );
    
    return( group );
}
