/**
* The MIT License (MIT)
* 
* Copyright (c) 2013 Retroswitch LLC
* 
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
* 
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
* 
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
* THE SOFTWARE.
*
*/

//
// Includes

#include "Flyer.h"
#include <cbm.h>
#include <conio.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

//
// Local Prototypes

void localTokenizeStatus( char* buf, char* toks[4] );
void localParseIpAddress( char* buf, uint8 ipAddress[4] );

//
// Functions

/******************************************************************************/
/* FlyerStartup                                                               */
/******************************************************************************/

int8 FlyerStartup()
{
    char buf[64];
    int rval;

    if( cbm_open( 7, 7, 15, "" ) != 0 )
    {
        return( -1 );
    }

    rval = cbm_read( 7, buf, 64 );

    if( rval < 2 || buf[0] != '0' || buf[1] != '0' )
    {
        return( -1 );
    }
    
    return( 0 );
}

/******************************************************************************/
/* FlyerShutdown                                                              */
/******************************************************************************/

void FlyerShutdown()
{
    cbm_close( 7 );
}

/******************************************************************************/
/* FlyerRequireFirmware                                                       */
/******************************************************************************/

int8 FlyerRequireFirmware( const char* version )
{
    char buf[64];
    char* toks[4];
    int rval;

    if( cbm_write( 7, "s:fi", 4 ) < 0 )
    {
        return( 0 );
    }

    rval = cbm_read( 7, buf, 64 );

    if( rval < 14 ) // minimum possible length
    {
        return( 0 );
    }

    buf[rval] = 0;

    toks[0] = buf;
    toks[1] = strchr( toks[0] + 1, ',' );
    toks[2] = strchr( toks[1] + 1, ',' );
    toks[3] = strchr( toks[2] + 1, ',' );
    toks[1][0] = 0; toks[1]++;
    toks[2][0] = 0; toks[2]++;
    toks[3][0] = 0; toks[3]++;

    // Treat error condition as unsupported firmware (status:firmware command not recognized)

    if( toks[0][0] != '0' && toks[0][1] != '0' )
    {
        return( 0 );
    }  

    // Compare version with a simple string compare

    return( strcmp( toks[1], version ) < 0 ? 0 : 1 );
}

/******************************************************************************/
/* FlyerListen                                                                */
/******************************************************************************/

int8 FlyerListen( uint8 channel, uint16 port )
{
    char buf[64];
    sprintf( buf, "listen:%d", port );

    if( cbm_open( channel, 7, channel, buf ) != 0 )
    {
        return( -1 );
    }

    if( cbm_read( 7, buf, 64 ) < 2 || buf[0] != '0' || buf[1] != '0' )
    {
        return( -1 );
    }

    return( 0 );
}

/******************************************************************************/
/* FlyerGetConnection                                                         */
/******************************************************************************/

int8 FlyerGetConnection( uint8 channel, uint8 remoteIp[4], uint16* remotePort )
{
    char buf[64];
    char* toks[4];
    int rval;

    sprintf( buf, "c-i:%d", channel );

    if( cbm_write( 7, buf, strlen( buf ) ) < 0 )
    {
        return( -1 );
    }

    rval = cbm_read( 7, buf, 64 );

    if( rval < 2 || buf[0] != '0' || buf[1] != '0' )
    {
        return( -1 );
    }

    buf[rval] = 0;

    localTokenizeStatus( buf, toks );

    if( atoi( toks[3] ) != 0 )
    {
        *remotePort = atoi( toks[2] );
        localParseIpAddress( toks[1], remoteIp );

        return( 1 );
    }

    return( 0 );
}

/******************************************************************************/
/* FlyerGetBytesAvailable                                                     */
/******************************************************************************/

int FlyerGetBytesAvailable( uint8 channel )
{
    char buf[64];
    char* toks[4];
    int rval;

    sprintf( buf, "n-s:%d", channel );

    if( cbm_write( 7, buf, strlen( buf ) ) < 0 )
    {
        return( -1 );
    }

    rval = cbm_read( 7, buf, 64 );

    if( rval < 2 || buf[0] != '0' || buf[1] != '0' )
    {
        return( -1 );
    }

    buf[rval] = 0;

    localTokenizeStatus( buf, toks );

    if( atoi( toks[3] ) == 0 )
    {
        return( -1 );
    }

    return( atoi( toks[2] ) );
}

/******************************************************************************/
/* FlyerGetIpAddress                                                          */
/******************************************************************************/

int8 FlyerGetIpAddress( uint8 flyerIp[4] )
{
    char buf[64];
    char* toks[4];
    int rval;

    if( cbm_write( 7, "s:ip", 4 ) < 0 )
    {
        return( -1 );
    }

    rval = cbm_read( 7, buf, 64 );

    if( rval < 2 || buf[0] != '0' || buf[1] != '0' )
    {
        return( -1 );
    }

    buf[rval] = 0;

    localTokenizeStatus( buf, toks );
    localParseIpAddress( toks[1], flyerIp );

    return( 0 );
}

//
// Local Functions

/******************************************************************************/
/* localTokenizeStatus                                                        */
/******************************************************************************/

void localTokenizeStatus( char* buf, char* toks[4] )
{
    toks[0] = buf;
    toks[1] = strchr( toks[0] + 1, ',' );
    toks[2] = strchr( toks[1] + 1, ',' );
    toks[3] = strchr( toks[2] + 1, ',' );
    toks[1][0] = 0; toks[1]++;
    toks[2][0] = 0; toks[2]++;
    toks[3][0] = 0; toks[3]++;
}

/******************************************************************************/
/* localParseIpAddress                                                        */
/******************************************************************************/

void localParseIpAddress( char* buf, uint8 ipAddress[4] )
{
    char* toks[4];

    toks[0] = buf;
    toks[1] = strchr( toks[0] + 1, '.' );
    toks[2] = strchr( toks[1] + 1, '.' );
    toks[3] = strchr( toks[2] + 1, '.' );
    toks[1][0] = 0; toks[1]++;
    toks[2][0] = 0; toks[2]++;
    toks[3][0] = 0; toks[3]++;

    ipAddress[0] = (uint8)atoi( toks[0] );
    ipAddress[1] = (uint8)atoi( toks[1] );
    ipAddress[2] = (uint8)atoi( toks[2] );
    ipAddress[3] = (uint8)atoi( toks[3] );
}

